"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * The unique id is used for unique hashes.
 */
var uniqueId = 0;
/**
 * Quick dictionary lookup for unit-less numbers.
 */
var CSS_NUMBER = Object.create(null);
/**
 * CSS properties that are valid unit-less numbers.
 *
 * Ref: https://github.com/facebook/react/blob/master/packages/react-dom/src/shared/CSSProperty.js
 */
var CSS_NUMBER_KEYS = [
    "animation-iteration-count",
    "border-image-outset",
    "border-image-slice",
    "border-image-width",
    "box-flex",
    "box-flex-group",
    "box-ordinal-group",
    "column-count",
    "columns",
    "counter-increment",
    "counter-reset",
    "flex",
    "flex-grow",
    "flex-positive",
    "flex-shrink",
    "flex-negative",
    "flex-order",
    "font-weight",
    "grid-area",
    "grid-column",
    "grid-column-end",
    "grid-column-span",
    "grid-column-start",
    "grid-row",
    "grid-row-end",
    "grid-row-span",
    "grid-row-start",
    "line-clamp",
    "line-height",
    "opacity",
    "order",
    "orphans",
    "tab-size",
    "widows",
    "z-index",
    "zoom",
    // SVG properties.
    "fill-opacity",
    "flood-opacity",
    "stop-opacity",
    "stroke-dasharray",
    "stroke-dashoffset",
    "stroke-miterlimit",
    "stroke-opacity",
    "stroke-width",
];
// Add vendor prefixes to all unit-less properties.
for (var _i = 0, CSS_NUMBER_KEYS_1 = CSS_NUMBER_KEYS; _i < CSS_NUMBER_KEYS_1.length; _i++) {
    var property = CSS_NUMBER_KEYS_1[_i];
    for (var _a = 0, _b = ["-webkit-", "-ms-", "-moz-", "-o-", ""]; _a < _b.length; _a++) {
        var prefix = _b[_a];
        CSS_NUMBER[prefix + property] = true;
    }
}
/**
 * Escape a CSS class name.
 */
function escape(str) {
    return str.replace(/[ !#$%&()*+,./;<=>?@[\]^`{|}~"'\\]/g, "\\$&");
}
/**
 * Interpolate the `&` with style name.
 */
function interpolate(selector, styleName) {
    return selector.replace(/&/g, styleName);
}
/**
 * Transform a JavaScript property into a CSS property.
 */
function hyphenate(propertyName) {
    return propertyName
        .replace(/[A-Z]/g, function (m) { return "-" + m.toLowerCase(); })
        .replace(/^ms-/, "-ms-"); // Internet Explorer vendor prefix.
}
/**
 * Generate a hash value from a string.
 */
function stringHash(str) {
    var value = 5381;
    var len = str.length;
    while (len--)
        value = (value * 33) ^ str.charCodeAt(len);
    return (value >>> 0).toString(36);
}
/**
 * Transform a style string to a CSS string.
 */
function styleToString(key, value) {
    if (value && typeof value === "number" && !CSS_NUMBER[key]) {
        return key + ":" + value + "px";
    }
    return key + ":" + value;
}
/**
 * Sort an array of tuples by first value.
 */
function sortTuples(value) {
    return value.sort(function (a, b) { return (a[0] > b[0] ? 1 : -1); });
}
/**
 * Stringify an array of property tuples.
 */
function stringifyProperties(properties) {
    return properties
        .map(function (_a) {
        var name = _a[0], value = _a[1];
        if (!Array.isArray(value))
            return styleToString(name, value);
        return value.map(function (x) { return styleToString(name, x); }).join(";");
    })
        .join(";");
}
/**
 * Interpolate CSS selectors.
 */
function child(selector, parent) {
    if (!selector)
        return parent;
    if (!selector.includes("&"))
        return parent + " " + selector;
    return interpolate(selector, parent);
}
/**
 * Recursive loop building styles with deferred selectors.
 */
function stylize(rulesList, stylesList, className, styles, parentClassName) {
    var properties = [];
    var nestedStyles = [];
    // Sort keys before adding to styles.
    for (var _i = 0, _a = Object.keys(styles); _i < _a.length; _i++) {
        var key_1 = _a[_i];
        var value = styles[key_1];
        if (key_1.charCodeAt(0) !== 36 /* $ */ && value != null) {
            if (typeof value === "object" && !Array.isArray(value)) {
                nestedStyles.push([key_1, value]);
            }
            else {
                properties.push([hyphenate(key_1), value]);
            }
        }
    }
    var isUnique = !!styles.$unique;
    var parent = styles.$global ? "" : parentClassName;
    var nested = parent ? nestedStyles : sortTuples(nestedStyles);
    var style = stringifyProperties(sortTuples(properties));
    var pid = style;
    if (className.charCodeAt(0) === 64 /* @ */) {
        var childRules = [];
        var childStyles = [];
        // Nested styles support (e.g. `.foo > @media`).
        if (parent && style) {
            childStyles.push({ selector: parent, style: style, isUnique: isUnique });
        }
        for (var _b = 0, nested_1 = nested; _b < nested_1.length; _b++) {
            var _c = nested_1[_b], name = _c[0], value = _c[1];
            pid += stylize(childRules, childStyles, name, value, parent);
        }
        // Add new rule to parent.
        rulesList.push({
            selector: className,
            rules: childRules,
            styles: childStyles,
            style: parent ? "" : style,
        });
    }
    else {
        var selector = parent ? child(className, parent) : className;
        if (style) {
            stylesList.push({ selector: selector, style: style, isUnique: isUnique });
        }
        for (var _d = 0, nested_2 = nested; _d < nested_2.length; _d++) {
            var _e = nested_2[_d], name = _e[0], value = _e[1];
            pid += stylize(rulesList, stylesList, name, value, selector);
        }
    }
    return pid;
}
/**
 * Transform `stylize` tree into style objects.
 */
function compose(cache, pid, rulesList, stylesList, name) {
    for (var _i = 0, stylesList_1 = stylesList; _i < stylesList_1.length; _i++) {
        var _a = stylesList_1[_i], selector = _a.selector, style = _a.style, isUnique = _a.isUnique;
        var key_2 = interpolate(selector, name);
        var item = new Style(style, isUnique ? "u\0" + (++uniqueId).toString(36) : "s\0" + pid + "\0" + style);
        item.add(new Selector(key_2, "k" + key_2));
        cache.add(item);
    }
    for (var _b = 0, rulesList_1 = rulesList; _b < rulesList_1.length; _b++) {
        var _c = rulesList_1[_b], selector = _c.selector, style = _c.style, rules = _c.rules, styles = _c.styles;
        var key_3 = interpolate(selector, name);
        var item = new Rule(key_3, style, "r\0" + pid + "\0" + key_3 + "\0" + style);
        compose(item, pid, rules, styles, name);
        cache.add(item);
    }
}
/**
 * Cache to list to styles.
 */
function join(arr) {
    var res = "";
    for (var i = 0; i < arr.length; i++)
        res += arr[i];
    return res;
}
/**
 * Implement a cache/event emitter.
 */
var Cache = /** @class */ (function () {
    function Cache(changes) {
        this.changes = changes;
        this.sheet = [];
        this.changeId = 0;
        this._keys = [];
        this._children = Object.create(null);
        this._counters = Object.create(null);
    }
    Cache.prototype.add = function (style) {
        var count = this._counters[style.id] || 0;
        var item = this._children[style.id] || style.clone();
        this._counters[style.id] = count + 1;
        if (count === 0) {
            this._children[item.id] = item;
            this._keys.push(item.id);
            this.sheet.push(item.getStyles());
            this.changeId++;
            if (this.changes)
                this.changes.add(item, this._keys.length - 1);
        }
        else if (item instanceof Cache && style instanceof Cache) {
            var prevItemChangeId = item.changeId;
            item.merge(style);
            if (item.changeId !== prevItemChangeId) {
                var index = this._keys.indexOf(style.id);
                this.sheet.splice(index, 1, item.getStyles());
                this.changeId++;
                if (this.changes)
                    this.changes.change(item, index, index);
            }
        }
    };
    Cache.prototype.remove = function (style) {
        var count = this._counters[style.id];
        if (count) {
            this._counters[style.id] = count - 1;
            var item = this._children[style.id];
            var index = this._keys.indexOf(item.id);
            if (count === 1) {
                delete this._counters[style.id];
                delete this._children[style.id];
                this._keys.splice(index, 1);
                this.sheet.splice(index, 1);
                this.changeId++;
                if (this.changes)
                    this.changes.remove(item, index);
            }
            else if (item instanceof Cache && style instanceof Cache) {
                var prevChangeId = item.changeId;
                item.unmerge(style);
                if (item.changeId !== prevChangeId) {
                    this.sheet.splice(index, 1, item.getStyles());
                    this.changeId++;
                    if (this.changes)
                        this.changes.change(item, index, index);
                }
            }
        }
    };
    Cache.prototype.values = function () {
        var _this = this;
        return this._keys.map(function (key) { return _this._children[key]; });
    };
    Cache.prototype.merge = function (cache) {
        for (var _i = 0, _a = cache.values(); _i < _a.length; _i++) {
            var item = _a[_i];
            this.add(item);
        }
        return this;
    };
    Cache.prototype.unmerge = function (cache) {
        for (var _i = 0, _a = cache.values(); _i < _a.length; _i++) {
            var item = _a[_i];
            this.remove(item);
        }
        return this;
    };
    Cache.prototype.clone = function () {
        return new Cache().merge(this);
    };
    return Cache;
}());
exports.Cache = Cache;
/**
 * Selector is a dumb class made to represent nested CSS selectors.
 */
var Selector = /** @class */ (function () {
    function Selector(selector, id) {
        this.selector = selector;
        this.id = id;
    }
    Selector.prototype.getStyles = function () {
        return this.selector;
    };
    Selector.prototype.clone = function () {
        return this;
    };
    return Selector;
}());
exports.Selector = Selector;
/**
 * The style container registers a style string with selectors.
 */
var Style = /** @class */ (function (_super) {
    __extends(Style, _super);
    function Style(style, id) {
        var _this = _super.call(this) || this;
        _this.style = style;
        _this.id = id;
        return _this;
    }
    Style.prototype.getStyles = function () {
        return this.sheet.join(",") + "{" + this.style + "}";
    };
    Style.prototype.clone = function () {
        return new Style(this.style, this.id).merge(this);
    };
    return Style;
}(Cache));
exports.Style = Style;
/**
 * Implement rule logic for style output.
 */
var Rule = /** @class */ (function (_super) {
    __extends(Rule, _super);
    function Rule(rule, style, id) {
        var _this = _super.call(this) || this;
        _this.rule = rule;
        _this.style = style;
        _this.id = id;
        return _this;
    }
    Rule.prototype.getStyles = function () {
        return this.rule + "{" + this.style + join(this.sheet) + "}";
    };
    Rule.prototype.clone = function () {
        return new Rule(this.rule, this.style, this.id).merge(this);
    };
    return Rule;
}(Cache));
exports.Rule = Rule;
/**
 * Generate class name from styles.
 */
function key(pid, styles) {
    var key = "f" + stringHash(pid);
    if (process.env.NODE_ENV === "production" || !styles.$displayName)
        return key;
    return styles.$displayName + "_" + key;
}
/**
 * The FreeStyle class implements the API for everything else.
 */
var FreeStyle = /** @class */ (function (_super) {
    __extends(FreeStyle, _super);
    function FreeStyle(id, changes) {
        var _this = _super.call(this, changes) || this;
        _this.id = id;
        return _this;
    }
    FreeStyle.prototype.registerStyle = function (css) {
        var ruleList = [];
        var styleList = [];
        var pid = stylize(ruleList, styleList, "", css, ".&");
        var id = key(pid, css);
        compose(this, pid, ruleList, styleList, 
        // Escape selector used in CSS.
        process.env.NODE_ENV === "production" ? id : escape(id));
        return id;
    };
    FreeStyle.prototype.registerKeyframes = function (keyframes) {
        return this.registerHashRule("@keyframes", keyframes);
    };
    FreeStyle.prototype.registerHashRule = function (prefix, styles) {
        var _a;
        return this.registerStyle((_a = {
                $global: true,
                $displayName: styles.$displayName
            },
            _a[prefix + " &"] = styles,
            _a));
    };
    FreeStyle.prototype.registerRule = function (rule, styles) {
        var _a;
        return this.registerStyle((_a = { $global: true }, _a[rule] = styles, _a));
    };
    FreeStyle.prototype.registerCss = function (styles) {
        return this.registerRule("", styles);
    };
    FreeStyle.prototype.getStyles = function () {
        return join(this.sheet);
    };
    FreeStyle.prototype.clone = function () {
        return new FreeStyle(this.id, this.changes).merge(this);
    };
    return FreeStyle;
}(Cache));
exports.FreeStyle = FreeStyle;
/**
 * Exports a simple function to create a new instance.
 */
function create(changes) {
    return new FreeStyle("f" + (++uniqueId).toString(36), changes);
}
exports.create = create;
//# sourceMappingURL=index.js.map